#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/un.h>
#include <stdlib.h>

// The loop count can also be given as argument
#define DEFAULT_COUNT 10000000

typedef struct thr_handle_t {
  pthread_mutex_t mutex;
  pthread_cond_t  cond;
  pthread_t       tid;
  int             stop;
  int             running;
} thr_handle;


pthread_t g_parent_tid=0;
int callback();

int init_thread_handle(thr_handle* h) {
  int ret = 0;
  thr_handle* handle = h;

  handle->tid = 0;
  handle->stop = 0;
  handle->running = 0;

  if((ret = pthread_mutex_init(&handle->mutex, NULL))<0) {
    fprintf(stderr, "Couldn't init new mutex lock (tid:%ld) %s:%d\n",
	    pthread_self(), __FILE__, __LINE__);
    return ret;
  }

  if(( ret = pthread_cond_init(&handle->cond, NULL))<0) {
    fprintf(stderr, "Couldn't init new condition variable: %s (tid:%ld) %s:%d\n",
            strerror(ret), pthread_self(), __FILE__, __LINE__);
    pthread_mutex_destroy(&handle->mutex);
    return ret;
  }

  return ret;
}


void* thread_proc(void* h) {
  int ret = 0;
  thr_handle *handle = (thr_handle *) h;

  if((ret = pthread_detach(pthread_self()))!=0) {
    fprintf(stderr, "Error when changing thread to detached: %s (tid:%ld) %s:%d\n",
	    strerror(errno), pthread_self(), __FILE__, __LINE__);
  }

  if((ret = pthread_mutex_lock(&handle->mutex))!=0) {
    fprintf(stderr, "Error when aquiring mutex: %s (tid:%ld) %s:%d\n",
	    strerror(ret), pthread_self(), __FILE__, __LINE__);
  }

  handle->running = 1;

  while(handle->stop==0 && ret == 0) {

    if((ret = pthread_cond_wait(&handle->cond, &handle->mutex))!=0) {
      fprintf(stderr, "Error when waiting signal: %s (tid:%ld) %s:%d\n",
	      strerror(ret), pthread_self(),__FILE__,__LINE__);
    }

    if((ret = pthread_mutex_unlock(&handle->mutex))!=0) {
      fprintf(stderr, "Error when releasing mutex: %s (tid:%ld) %s:%d\n",
	      strerror(ret), pthread_self(), __FILE__, __LINE__);
    }

    if(ret == 0) {
      ret = callback();
    }

    if((ret = pthread_mutex_lock(&handle->mutex))!=0) {
      fprintf(stderr, "Error when aquiring mutex: %s (tid:%ld) %s:%d\n",
	      strerror(ret), pthread_self(), __FILE__, __LINE__);
    }

    if(ret != 0)
      break;
  }


  if((ret = pthread_mutex_unlock(&handle->mutex))!=0) {
    fprintf(stderr, "Error when releasing mutex: %s (tid:%ld) %s:%d\n",
	    strerror(ret), pthread_self(), __FILE__, __LINE__);
  }

  if((ret = pthread_mutex_destroy(&handle->mutex))!=0) {
    fprintf(stderr, "Error when destroying mutex: %s (tid:%ld) %s:%d\n",
	    strerror(ret), pthread_self(), __FILE__, __LINE__);
  }

  if((ret = pthread_cond_destroy(&handle->cond))!=0) {
    fprintf(stderr, "Error when destroying condvar: %s (tid:%ld) %s:%d\n",
	    strerror(ret), pthread_self(), __FILE__, __LINE__);
  }

  // Make sure that we don't try to free the parent's handle since it is in stack
  // of main.
  if(handle!=NULL && !pthread_equal(g_parent_tid, handle->tid) ) {
    free(handle);
    handle=NULL;
  }
  return (void*)ret;
}


int start_thread(thr_handle* h) {
  int ret = 0;
  thr_handle* handle = h;

  if((ret = pthread_create(&handle->tid, NULL, thread_proc, (void*) handle) )!=0) {
    fprintf(stderr, "Couldn't create a new thread: %s (tid:%ld) %s:%d\n",
            strerror(ret), pthread_self(), __FILE__, __LINE__);
    return ret;
  }

  return ret;
}


int stop_thread(thr_handle* h) {
  int ret = 0;
  thr_handle* handle = h;

  if((ret = pthread_mutex_lock(&handle->mutex))!=0) {
    fprintf(stderr, "Couldn't create a new thread: %s (tid:%ld) %s:%d\n",
            strerror(ret), pthread_self(), __FILE__, __LINE__);
    return ret;
  }

  handle->stop = 1;

  if((ret = pthread_cond_signal(&handle->cond))!=0) {
    fprintf(stderr, "Couldn't signal condvar: %s (tid:%ld) %s:%d\n",
            strerror(ret), pthread_self(), __FILE__, __LINE__);
    return ret;
  }

  if((ret = pthread_mutex_unlock(&handle->mutex))!=0) {
    return ret;
  }

  return ret;
}


int callback()
{
  int status = 0;

  thr_handle* thr_data = NULL;

  /* This is freed on exit from thread_proc */
  if((thr_data = (thr_handle*)malloc(sizeof(thr_handle)))== NULL){
    fprintf(stderr, "Couldn't allocate memory for new thr_handle: %s (tid:%ld) %s:%d\n",
	    strerror(errno), pthread_self(), __FILE__, __LINE__);
    return -1;
  }

  if((status = init_thread_handle(thr_data))==0)
    status = start_thread(thr_data);
  else
    fprintf(stderr, "Couldn't init new thread (tid:%ld err:%s) %s:%d\n",
	    pthread_self(), strerror(errno), __FILE__, __LINE__);

  /* Do something that causes delay */
#ifdef DO_LAG
  struct timeval tv;
  tv.tv_sec = 0;
  tv.tv_usec = 100000;
  select(0, NULL, NULL, NULL, &tv);
#endif

  if(thr_data != NULL)
    status = stop_thread(thr_data);

  return status;
}

/*
  da main
 */
int main( int argc, char** argv)
{
  int ret = 0, i=0;
  int count = 0;
  thr_handle parent_thread;

  if(argc == 2) {
    if((ret = sscanf(argv[1], "%d", &count))==0) {
      fprintf(stderr, "Enter an integer as argument in order to define loop count\n");
      exit(1);
    }
  }
  else
    count = DEFAULT_COUNT;

  printf("Looping %d times\n", count);

  if((ret = init_thread_handle(&parent_thread))!=0) {
    fprintf(stderr, "Failed to init parent thread (%s)\n",
	    strerror(errno));
  }
  else if((ret = start_thread(&parent_thread))!=0) {
    fprintf(stderr, "Failed to start parent thread (%s)\n",
	    strerror(errno));
  }
  if(ret==0) {
    g_parent_tid = parent_thread.tid;
  }

  for(i=0; i<count; i++)
    {
    if((ret = pthread_cond_signal(&parent_thread.cond))!=0) {
      fprintf(stderr, "Failed to signal parent thread from main: %s\n",
              strerror(errno));
      break;
    }
  }

  if((ret = stop_thread(&parent_thread))!=0) {
    fprintf(stderr, "Failed to stop parent thread (%s)\n",
	    strerror(errno));
  }

  /*** Clean up ***/
  ret = stop_thread(&parent_thread);
  pthread_join(parent_thread.tid, NULL);

  pthread_cond_destroy(&parent_thread.cond);
  pthread_mutex_destroy(&parent_thread.mutex);

  fflush(stderr);

  fprintf(stderr, "EXITING\n");

  return ret;
}
